/*
 * driver.c
 *
 * driver functions for pc
 */


#include <windows.h> 

#include <math.h>
#include "lightf.h"
#include "global.h"

#define ALIGNLONG(i) ((i+3)/4*4)
/* some globals */

extern float xobs, yobs, zobs;
extern float xcenter;
extern float ycenter;
extern float zcenter;
extern float oldxobs, oldyobs, oldzobs;
extern float oldxcenter, oldycenter, oldzcenter;

extern float fov;
extern long xsize, ysize;
extern HWND globalWinHandle;
extern int frameTimeMode;
extern int matrixTimeMode;
extern int zoomRate;
extern int leftdown, middledown, rightdown;
extern int autoRefine;

static int lastzoomRate=2;

#define PI 3.1415926
#define FILTER_FLAG 0x00ffffff
#define RED_FLAG    0X000000ff
#define GREEN_FLAG  0x0000ff00
#define BLUE_FLAG   0X00ff0000
#define REFINE_ZOOM 0x2

typedef char Array3[3];

/* global bitmap information */  
static BYTE BitMapBits[WEI_MAX_WINDOW_SIZE*WEI_MAX_WINDOW_SIZE*3];
static BITMAPINFO bminfo;  
/* function to get current transformation information */

void
pc_get_transform(Matrix4 mat_model, Matrix4 mat_proj,
	int *x, int *y, int *width, int *height)
{
  void pcSetIdentity(Matrix4 in);
  void pcOutputMatrix(Matrix4 in);
  void pcFindModelMatrix(Matrix4 m);
  void pcFindProjectionMatrix(Matrix4 p);
  DWORD oldTime, elapsedTime;
  char temp[200];
  HDC hDC;

  *x = 0;
  *y = 0;

  if( (autoRefine&REFINE_ZOOM) && !(leftdown || middledown || rightdown) )
    {
      *width = xsize;
      *height = ysize;
      
    }
  else
    {
      *width = xsize/zoomRate;
      *height = ysize/zoomRate;
    }
  
  pcFindModelMatrix(mat_model);
  
  pcFindProjectionMatrix(mat_proj);

}


/* function to draw stuff into current framebuffer */



/* drive that strech the image twice */
void
pc_write_pixrect_twice(int x, int y, int w, int h,
	int format, int type, void *pixels)
{
  
  int i,j;
  PAINTSTRUCT paintStruct;
  HDC hDCforPainting;
  COLORREF paintColor;
  HDC hDC, hMemDC, anotherhDC;
  int offset1, offset2;
  unsigned char  *pBuffer;
  char temp[200];
  int pad_bytes;
  
  /* bitmap stuff */
  static HBITMAP hBitmap, anotherhBitmap;
 
 
  static int firstTime=1;
  DWORD oldTime, elapsedTime;
  void drawPreviousFrame(void);

  /* code */
  if( firstTime )
    {
      firstTime = 0;
      
      /* intitialize bitmap info header */
      bminfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
      bminfo.bmiHeader.biWidth = w;
      bminfo.bmiHeader.biHeight = h;
      bminfo.bmiHeader.biPlanes = 1;
      bminfo.bmiHeader.biBitCount = 24;
      bminfo.bmiHeader.biCompression = BI_RGB;
      bminfo.bmiHeader.biSizeImage = ALIGNLONG(w*3)*h;
      bminfo.bmiHeader.biXPelsPerMeter = 0;
      bminfo.bmiHeader.biYPelsPerMeter = 0;
      bminfo.bmiHeader.biClrUsed = 0;
      bminfo.bmiHeader.biClrImportant = 0;

    }


  bminfo.bmiHeader.biWidth = w;
  bminfo.bmiHeader.biHeight = h;
  bminfo.bmiHeader.biSizeImage = ALIGNLONG(w*3)*h;
  pBuffer = pixels;
  offset1 = 0;
  pad_bytes = ALIGNLONG(w*3) - w*3;

  
  hDC = GetDC(globalWinHandle);
  hMemDC = CreateCompatibleDC(hDC);
  hBitmap = CreateDIBitmap(hDC, &bminfo.bmiHeader, CBM_INIT, 
			   BitMapBits, &bminfo, DIB_RGB_COLORS);
  

  SelectObject(hMemDC, hBitmap);
  
  offset2 = 0;

  for(j=y+h-1; j>=y; j--)
    {
      for(i=x; i<=x+w-1; i++)
	{
	  /*
	  offset2 = ((i-x)+(y+h-1-j)*w)*3;
	  */
	  
	  offset1++;

	  BitMapBits[offset2++] = pBuffer[offset1++];
	  BitMapBits[offset2++] = pBuffer[offset1++];
	  BitMapBits[offset2++] = pBuffer[offset1++];
	  
	}
      offset2 += pad_bytes;
    }

  if( matrixTimeMode)
    {
      oldTime = GetCurrentTime();
    }

  if( xsize/w == lastzoomRate )
    {
      StretchBlt(hDC, 0, 0, xsize, ysize, hMemDC, 0, 0, w, h, SRCCOPY);
    
    }
  else
    {
      drawPreviousFrame();
    }

  lastzoomRate = xsize/w;

  ReleaseDC(globalWinHandle, hDC);
  DeleteDC(hMemDC);
  DeleteObject(hBitmap);
 
}



/* install the drivers */
LFDriver pc_driver = {
    pc_get_transform,
    pc_write_pixrect_twice, 
    /* irisgl_write_pixrect_smooth, */
};

LFDriver *__lf_driver = &pc_driver;

/* function that sets a matrix to be identity matrix */
void pcSetIdentity(Matrix4 in)
{
  int i,j;

  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      {
	in[i][j] = 0;
      }

  for(i=0; i<4; i++)
    {
      in[i][i] = 1;
    }
}

/* function to output a matrix */
void pcOutputMatrix(Matrix4 in)
{
  int i,j;
  char temp[200];

  for(i=0; i<4; i++)
    {
      
      sprintf(temp, "%f, %f, %f, %f", in[i][0], in[i][1], in[i][2], in[i][3] );
    
    }
}


/* function to calculate model matrix from geometry */
void pcFindModelMatrix(Matrix4 m)
{
  float zx, zy, zz;
  float xx, xy, xz;
  float yx, yy, yz;
  float A, B, C; /* the magnitude of vector z and x , y*/
  float delta;

  /* find z vector */
  zx = xobs - xcenter;
  zy = yobs - ycenter;
  zz = zobs - zcenter;
  A = sqrt( zx*zx + zy*zy + zz*zz );
  if( A == 0 )
    {
      printf("xobs:%f, yobs:%f, zobs:%f\n", xobs, yobs, zobs);
      printf("xcenter:%f, ycenter:%f, zcenter:%f\n", xcenter, ycenter, zcenter);
      printf("zx:%f, zy:%f, zz:%f\n", zx, zy, zz);

      lfError("weiFindModelMatrix(): A is zero\n");
      exit(16);
    }
  else
    {
      zx /= A;
      zy /= A;
      zz /= A;
    }

  /* find x vector */
  xx = zobs - zcenter;
  xy = 0;
  xz = -xobs + xcenter;
  B = sqrt( xx*xx + xy*xy + xz*xz );
  if( B == 0 )
    {
      lfError("weiFiondModelMatrix(): B is zero\n");
      exit(17);
    }
  else
    {
      xx /= B;
      xy /= B;
      xz /= B;
    }

  /* find y */
  yx = zy*xz;
  yy = zz*xx - zx*xz;
  yz = -zy*xx;
  C = sqrt(yx*yx + yy*yy + yz*yz);
  if( C == 0 )
    {
      lfError("weiFindModelMatrix(): C is zero\n");
      exit(18);
    }
  else
    {
      yx /= C;
      yy /= C;
      yz /= C;
    }

  /* find model matrix */
  delta = xx*(yy*zz-yz*zy)-xy*(yx*zz-zx*yz)+xz*(yx*zy-zx*yy);
  
  if( delta == 0 )
    {
      lfError("weiFindModelMatrix(): delta is zero\n");
      exit(18);
    }
  else
    {
      delta = 1.0/delta;

      m[0][0] = (yy*zz-zy*yz)*delta;
      m[0][1] = -(yx*zz-yz*zx)*delta;
      m[0][2] = (yx*zy-zx*yy)*delta;
      m[0][3] = -xobs*m[0][0]-yobs*m[0][1]-zobs*m[0][2];

      m[1][0] = -(xy*zz-zy*xz)*delta;
      m[1][1] = (xx*zz-xz*zx)*delta;
      m[1][2] = -(xx*zy-xy*zx)*delta;
      m[1][3] = -xobs*m[1][0]-yobs*m[1][1]-zobs*m[1][2];

      m[2][0] = (xy*yz-xz*yy)*delta;
      m[2][1] = -(xx*yz-xz*yx)*delta;
      m[2][2] = (xx*yy-xy*yx)*delta;
      m[2][3] = -xobs*m[2][0]-yobs*m[2][1]-zobs*m[2][2];

      m[3][0] = 0;
      m[3][1] = 0;
      m[3][2] = 0;
      m[3][3] = 1;
    }
} 


/* find out the projection matrix */
void pcFindProjectionMatrix(Matrix4 p)
{
  float value;
  int i,j;

  value = 1.0/tan(fov/2*PI/180.0);

  for(i=0; i<4; i++)
    for(j=0; j<4; j++)
      {
	p[i][j] = 0;
      }
  
  p[0][0] = value;
  p[1][1] = value;
  p[2][2] = -1;
  p[3][2] = -1;
  p[2][3] = -0.002;
}

/* draw the previous frame */
void drawPreviousFrame(void)
{
  HDC hDC, hMemDC;
  static HBITMAP hBitmap;

  hDC = GetDC(globalWinHandle);
  hMemDC = CreateCompatibleDC(hDC);
  hBitmap = CreateDIBitmap(hDC, &bminfo.bmiHeader, CBM_INIT, 
			   BitMapBits, &bminfo, DIB_RGB_COLORS);
  SelectObject(hMemDC, hBitmap);

  /* draw here */
  StretchBlt(hDC, 0, 0, xsize, ysize, hMemDC, 0, 0, 
	     bminfo.bmiHeader.biWidth, bminfo.bmiHeader.biHeight , SRCCOPY);

  ReleaseDC(globalWinHandle, hDC);
  DeleteDC(hMemDC);
  DeleteObject(hBitmap);
}








