/*
 *  libstubs.c: Stub routines for SGI libraries.
 *  Many functions are not needed for a viewer that only reads
 *  .lif files.  Thus, we use this file to define the functions
 *  (so that the compiler can link it). None of these functions
 *  should ever be called within the viewer.
 *
 */

#include <stdio.h>
#include <stdlib.h>


/* **********************************************************************
 *
 *			NO VQ LIBRARY
 *
 * **********************************************************************
 */

#ifdef _NO_VQ_LIBRARY_

/* If we only read .lif files, and don't do any compression in the
 * viewer, then we don't need the vq library....
 * Thus we define the functions as dummy functions, since they will
 * never be called.
 */

#include "vq_iface.h"

VQ_t *vqGenTreeCodebook(const unsigned char *data_array, int data_size,
	int dw_size, int max_code_size, int cw_size) 
{
  return NULL;
}

void vqEncode(VQ_t *vq, unsigned char *image, int domain_dimens,
	int len[], int clen, int padded, int padded_nbits,
	int *codearray, int *bitoffset_p)
{
  return;
}

VQ_t *vqTreeCodebookToLinearCodebook(const VQ_t *tree_vq)
{
  return NULL;
}

void vqFreeCodebook(VQ_t *vq)
{
  return;
}

#endif /* _NO_VQ_LIBRARY_ */


/* **********************************************************************
 *
 *			NO IMAGE LIBRARY
 *
 * **********************************************************************
 */

#ifdef _NO_IMAGE_LIBRARY_

/* If we only read .lif files, and don't do any compression in the
 * viewer, then we don't need the IMAGE library....
 * Thus we define the functions as dummy functions, since they will
 * never be called.
 */

#include <gl/image.h>

IMAGE *iopen()
{
  return NULL;
}

int iclose(IMAGE *image)
{
  return 0;
}

int getrow(IMAGE *image, unsigned short *buffer, unsigned int y, 
	   unsigned int z)
{
  return 0;
}


#endif /* _NO_IMAGE_LIBRARY_ */



/* **********************************************************************
 *
 *			NO GEN LIBRARY
 *
 * **********************************************************************
 */

#ifdef _NO_GEN_LIBRARY_

char *dirname(char *path)
{
  /* This is the "easy" thing to return, even though this function
   * should never get called anyways...
   */
  return(".");
}


#endif /* _NO_GEN_LIBRARY_ */
