/*
 *  view.c, PC version 1.0alpha1.
 *    View a light field.
 *
 */

#include <windows.h>
#include <stdio.h>
#include "menu.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include "lightf.h"
#include "global.h"

#define PI 3.1415926

/* global property */
#define WINDOW_WIDTH 192
#define WINDOW_HEIGHT 192
static char szAppName[] = "viewer";
static char menuName[] = "generic";

HWND globalWinHandle = 0;

#define FZ 0.0
#define EPSILON 0.001

#define IMAGESIZEX 640
#define IMAGESIZEY 640

#define BASENAME "."

/*
Matrix Identity =
	{ 1, 0, 0, 0,  0, 1, 0, 0,  0, 0, 1, 0,  0, 0, 0, 1 };
	*/

#define MENU_POINT	1
#define MENU_UVINTER	2
#define MENU_STINTER	3
#define MENU_EXIT	4

static char menustr[] =
    "Options %t|Point|UV Interpolate|ST Interpolate %l|Exit";
static char globaltestfilename[] = "soda1.lif";

#define RESETXOBS 0.0
#define RESETYOBS 0.0
#define RESETZOBS 2.1

#define INITXCENTER 0.0
#define INITYCENTER 0.0
#define INITZCENTER 0.0

#define	UVLERP	0x1
#define	STLERP	0x2

#define REFINE_INTERP 0x1
#define REFINE_ZOOM   0x2
#define REFINE_UV 0x4

#define SIGN(x) (((x)<0) ? -1 : 1)

float xobs = RESETXOBS, yobs = RESETYOBS, zobs = RESETZOBS;
float xcenter = INITXCENTER;
float ycenter = INITYCENTER;
float zcenter = INITZCENTER;
float oldxobs, oldyobs, oldzobs, oldxcenter, oldycenter, oldzcenter;
float fov = 40;
long xsize = WINDOW_WIDTH, ysize = WINDOW_HEIGHT;
/*long xsize = 192, ysize = 192; */
int shift = 0;

/* zoom rate */
int zoomRate = 2;
 
/* refine or not */
int autoRefine = REFINE_INTERP ;

/* Local helper functions */
void loadview();

/* globals copied directly from main in view.c */
long xorig=0, yorig=0;
float lx=0, ly=0, mx=0, my=0, dx=0, dy=0; 
float rold=0, rnew=0;
float timing=0;
short val=0;
long dev=0;
int leftdown = 0, middledown = 0, rightdown = 0;
int updown=0, downdown=0;
int done = 0;
long menuid=0;
int interp = UVLERP;
char *basename=0;
int redraw = 1;
int frame = 0;
int single_frame = 0;
float clamp_ratio=0;
int i=0;
float newxobs=0, newyobs=0, newzobs=0;
float newxcenter=0, newycenter=0, newzcenter=0;
char *suffixPtr=0;
float draw_descr[] = {
  LF_DRAW_MAG_FILTER, LF_BILINEAR_UV_NEAREST_ST,
  LF_NULL,
};
float read_descr[] = {
  LF_NULL,
};

int timeMode = 0;
int frameTimeMode = 0;
int matrixTimeMode = 0;
int keyboardSpeed = 0;
int keyboardDefaultSpeed = 31;

DWORD oldTime, elapsedTime;

/* This is used as the temporary file if we need to gunzip it. */
char tempLif[1000] = "/tmp/current.lif";

typedef enum {
  LF_LID,
  LF_LIF,
  LF_LIFBADVER,
  LF_GZ,
  LF_UNKNOWN
} LF_FileType;

/*
// function prototypes 
*/

/*the handle for thw main window */
static HWND MainWindow=0;

/*call back routine for the main window
 */
LRESULT CALLBACK WndProc(HWND hWindow,
                         UINT Message,
                         WPARAM wParam,
                         LPARAM lParam
                         );

/*function to create the window
 */
HWND Create(HINSTANCE hInst, 
            int nCmdShow
            );
/*register */
BOOL Register(HINSTANCE hInst);

/*main */
int WINAPI WinMain(HINSTANCE hInst, 
                   HINSTANCE hPrevInstance,
                   LPSTR lpszCmdParam,
                   int nCmdShow 
                   );

/* WM_DESTROY handler */
void Window1_OnDestroy(HWND hwnd);

/* usage routine */
void PrintUsage(char *programName);
void ParseArgs(int argc, char argv[][200]);
LF_FileType DetermineFileType(char *fileName);

/*
// real main routine
*/

int WINAPI WinMain(HINSTANCE hInst,
                   HINSTANCE hPrevInstance,
                   LPSTR lpszCmdParam,
                   int nCmdShow
                   )
{
/* the message passed to the main window for all events
 */
   MSG Msg;
   void weiSetDeviceMode(void);
   int argc;
   char argv[10][200];
   int i, offsetForCmd;
   char temp[200];
   char *progName = argv[0];
   LF_FileType fileType = LF_UNKNOWN;
   
   /* functions */ 
   void handleCompressedFile(char *fileName);
   LF_FileType DetermineFileType(char *fileName);

   if(!hPrevInstance)
     {
       if( !Register(hInst) )
	 {
	   return FALSE;
        }
     }

   /* find out argc and argv */
   /*
   strcpy(argv[0], "pcviewer.exe");
   */

   for(i=1, argc=1, offsetForCmd=0; i<10; i++, argc++)
     {
       if( sscanf( ((char *)lpszCmdParam)+offsetForCmd, "%s", argv[i] ) == 1 )
	 {
	   offsetForCmd += strlen(argv[i]);
	   if( isspace( *( (char *)(((char *)lpszCmdParam)+offsetForCmd) ) ) )
	     {
	       offsetForCmd++;
	     }
	 } 
       else
	 {
	   break;
	 }
     }

  
   /* parse arguments */
   
   ParseArgs( argc, argv);

   /* set window size */
   xsize *= zoomRate;
   ysize *= zoomRate;

  /* Set window size (if environment variable specifies it) */
  if  (getenv("LF_WIN_SIZE")) {
    sscanf(getenv("LF_WIN_SIZE"), "%d%d", &xsize, &ysize);
  }

  /* Determine the type of file:
   * .lif, .lid, .gz, or other...
   */
  fileType = DetermineFileType(basename);

  /* gunzip, if necessary */
  if (fileType == LF_GZ) 
    {
      handleCompressedFile(basename);
      strcpy(basename, "temp.lif");
    }

  else if (fileType == LF_LIF) 
    {
      /* Check to make sure that file has proper .lif extension */
      /*
      suffixPtr = (strlen(basename) > 4) ? &(basename[strlen(basename)-4]) :
	basename;
	*/
      /* If it doesn't end in .lif, print warning message, don't abort */
      /*
      if (strcmp(".lif", suffixPtr) && strcmp(".LIF", suffixPtr)) {
	fprintf(stderr, "Warning:  data file %s does not have .lif"
		"extension!\n", basename);
	exit(45);
      }
      */
    }
  else 
    {
      exit(-1);
    }

  /* Get light field dataflow stuff ready */
  lfInitContext();

  /* Read in the file using new .lif file format */
  printf("Reading light field file %s\n", basename);
  lfBegin(LF_FIELD, NULL);
  lfRead(LF_LIGHTFIELD, read_descr, basename);
  lfEnd();

  system("del temp.lif");

  /*create the main window */ 
  globalWinHandle = -1;
  MainWindow = Create(hInst, nCmdShow);
  globalWinHandle = MainWindow;

  if(!MainWindow)
    {
      return FALSE;

    }
  
 
  /*loop around to handle messages */
  while( GetMessage(&Msg, NULL, 0, 0) )
    {
      TranslateMessage(&Msg);
      DispatchMessage(&Msg);
    }

/*return and exit */
  return Msg.wParam;
}


/*
// register the window
*/
BOOL Register(HINSTANCE hInst)
{
  WNDCLASS WndClass;

  /*assign attributes to the class */
  WndClass.style = CS_HREDRAW|CS_VREDRAW;
  WndClass.lpfnWndProc = WndProc; 
  WndClass.cbClsExtra = 0;
  WndClass.cbWndExtra = 0;
  WndClass.hInstance = hInst;
  WndClass.hIcon = LoadIcon(hInst, IDI_APPLICATION);
  WndClass.hCursor = LoadCursor(NULL, IDC_ARROW);
  WndClass.hbrBackground = GetStockObject(BLACK_BRUSH);
  WndClass.lpszMenuName = menuName; 
  WndClass.lpszClassName = szAppName; 

  
  return( RegisterClass(&WndClass) != 0 );
}


/*
// create the window 
*/
HWND Create(HINSTANCE hInstance, int nCmdShow)
{
  HWND hWindow = CreateWindow(
                                szAppName,
                                szAppName,
                                WS_OVERLAPPEDWINDOW,
                                CW_USEDEFAULT,
				CW_USEDEFAULT,
                                xsize, 
                                ysize+30,
                                NULL,
                                NULL,
                                hInstance,
                                NULL
                                );

  if( hWindow == NULL )
    {
      return hWindow;
    }

  /*display the window */
  ShowWindow(hWindow, nCmdShow);
  UpdateWindow(hWindow);

  return hWindow;
}


/*
// callback handler for main window
*/
LRESULT CALLBACK WndProc(HWND hWindow,
                         UINT Message,
                         WPARAM wParam,
                         LPARAM lParam
                         )
{
 /*first time enter this function? */
  static int firstTime = 1;
  static char buffer[200];
  int result;
  static int times = 0;
  int oldlerp;
  static UINT oldMessage = -1;
  UINT realOld;
  MSG peekMessage;

  /*variables for menu */
  HMENU hMenu;

  /* variables for mouse */ 
  static POINT pCursorOld, pCursorNew;
  POINTS pCursor;
  DWORD dwMesPos;

  /* variables for painting */
  PAINTSTRUCT paintStruct;
  HDC hDCforPainting;
  COLORREF paintColor;
  HDC hDC;

  /* variable for size */
  RECT rWindow;

  /* variables for the time */
  static DWORD lastDrawTime=0;
  
  /* variables for debug */
  int whichCase;

  /* my own message */
  static WORD wNewMessage = WM_NULL;

  /* any menu item selected? */
  static int anyItemSelected=1;

  /* fucntions */
  void drawPreviousFrame(void);
  LF_FileType DetermineFileType(char *fileName);

  /* initialize the mouse position */
  times ++;
  realOld = oldMessage;
 
  if( firstTime )
    {
      firstTime = 0;
      GetCursorPos(&pCursorOld);
      hMenu = GetMenu(hWindow);
      CheckMenuItem(hMenu, IDM_POP2, MF_BYCOMMAND|MF_CHECKED);
      CheckMenuItem(hMenu, IDM_POP6, MF_BYCOMMAND|MF_CHECKED);
      interp = UVLERP;

      /* refine mode stuff */
      if( autoRefine == 0 )
	{
	  CheckMenuItem(hMenu, IDM_POP9, MF_BYCOMMAND|MF_CHECKED);
	}
      else 
	{
	  if(autoRefine & REFINE_INTERP)
	    {
	      CheckMenuItem(hMenu, IDM_POP10, MF_BYCOMMAND|MF_CHECKED);
	    }
	  if(autoRefine & REFINE_ZOOM)
	    {
	      CheckMenuItem(hMenu, IDM_POP11, MF_BYCOMMAND|MF_CHECKED);
	    }
	  if(autoRefine & REFINE_UV)
	    {
	      CheckMenuItem(hMenu, IDM_POP12, MF_BYCOMMAND|MF_CHECKED);
	    }
	}

      SystemParametersInfo(SPI_GETKEYBOARDSPEED, 0, &keyboardDefaultSpeed, NULL);
    }

  
  /* always redraw */
  redraw = 1;

  if( Message == WM_PAINT )
    {
      switch(oldMessage)
	{
	case WM_PAINT:
	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	  
	  oldMessage = Message;
	  return;
	  break;
	default:
	  oldMessage = Message;
	  break;
	}
    }
  else
    {
      oldMessage = Message;
    }

  switch(Message)
    {
      /*destroy signal*/
    case WM_DESTROY:
      PostQuitMessage(0);
      break;

      /* size has changed */
    case WM_SIZE:
      
      GetWindowRect(hWindow, &rWindow);
      xsize = rWindow.right-rWindow.left;
      if( xsize > WEI_MAX_WINDOW_SIZE )
	{
	  xsize = WEI_MAX_WINDOW_SIZE;
	}
      
      ysize = rWindow.bottom-rWindow.top-30;
      if( ysize > WEI_MAX_WINDOW_SIZE )
	{
	  ysize = WEI_MAX_WINDOW_SIZE;
	}
   
      break;
      
	
      /*menu signal*/
    case WM_COMMAND:
      switch(wParam)
	{
	case IDM_POP1:
	  hMenu = GetMenu(hWindow);
	  
	  if( GetMenuState(hMenu, IDM_POP1, MF_BYCOMMAND) == MF_UNCHECKED )
	    {
	      CheckMenuItem(hMenu, IDM_POP1, MF_BYCOMMAND|MF_CHECKED);
	      CheckMenuItem(hMenu, IDM_POP2, MF_BYCOMMAND|MF_UNCHECKED);
	      CheckMenuItem(hMenu, IDM_POP3, MF_BYCOMMAND|MF_UNCHECKED);
	      
	      interp = 0;
	    }
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
	  break;
	
	case IDM_POP2:
	  hMenu = GetMenu(hWindow);
	  if( GetMenuState(hMenu, IDM_POP2, MF_BYCOMMAND) == MF_UNCHECKED )
	    {
	      CheckMenuItem(hMenu, IDM_POP2, MF_BYCOMMAND|MF_CHECKED);
	      CheckMenuItem(hMenu, IDM_POP1, MF_BYCOMMAND|MF_UNCHECKED);
	    }
	  else if( GetMenuState(hMenu, IDM_POP3, MF_BYCOMMAND) == MF_CHECKED )
	    {
	      CheckMenuItem(hMenu, IDM_POP2, MF_BYCOMMAND|MF_UNCHECKED);
	    }
	  else
	    {
	      CheckMenuItem(hMenu, IDM_POP2, MF_BYCOMMAND|MF_UNCHECKED);
	      CheckMenuItem(hMenu, IDM_POP1, MF_BYCOMMAND|MF_CHECKED);
	    }

	  if( interp & UVLERP )
	    interp &= ~UVLERP;
	  else
	    interp |= UVLERP;
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
	  break;
	
	case IDM_POP3:

	  hMenu = GetMenu(hWindow);
	  if( GetMenuState(hMenu, IDM_POP3, MF_BYCOMMAND) == MF_UNCHECKED )
	    {
	      CheckMenuItem(hMenu, IDM_POP3, MF_BYCOMMAND|MF_CHECKED);
	      CheckMenuItem(hMenu, IDM_POP1, MF_BYCOMMAND|MF_UNCHECKED);
	    }
	  else if( GetMenuState(hMenu, IDM_POP2, MF_BYCOMMAND) == MF_CHECKED )
	    {
	      CheckMenuItem(hMenu, IDM_POP3, MF_BYCOMMAND|MF_UNCHECKED);
	    }
	  else
	    {
	       CheckMenuItem(hMenu, IDM_POP3, MF_BYCOMMAND|MF_UNCHECKED);
	       CheckMenuItem(hMenu, IDM_POP1, MF_BYCOMMAND|MF_CHECKED);
	    }
	  if( interp & STLERP )
	      interp &= ~STLERP;
	    else
	      interp |= STLERP;
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
	  break;

	  /* reset menu item */
	case IDM_POP4:
	  xobs=RESETXOBS;
	  yobs=RESETYOBS;
	  zobs=RESETZOBS;
	  xcenter = INITXCENTER;
	  ycenter = INITYCENTER;
	  zcenter = INITZCENTER;
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
	
	  break;

	  
	case IDM_POP5:
	  hMenu = GetMenu(hWindow);
	  CheckMenuItem(hMenu, IDM_POP5, MF_BYCOMMAND|MF_CHECKED);
	  CheckMenuItem(hMenu, IDM_POP6, MF_BYCOMMAND|MF_UNCHECKED);
	  CheckMenuItem(hMenu, IDM_POP7, MF_BYCOMMAND|MF_UNCHECKED);
	  
	  zoomRate = 1;
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
       
	  break;

	case IDM_POP6:
	  hMenu = GetMenu(hWindow);
	  CheckMenuItem(hMenu, IDM_POP6, MF_BYCOMMAND|MF_CHECKED);
	  CheckMenuItem(hMenu, IDM_POP5, MF_BYCOMMAND|MF_UNCHECKED);
	  CheckMenuItem(hMenu, IDM_POP7, MF_BYCOMMAND|MF_UNCHECKED);
      
	  zoomRate = 2;
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
      
	  break;

	case IDM_POP7:
	  hMenu = GetMenu(hWindow);
	  CheckMenuItem(hMenu, IDM_POP7, MF_BYCOMMAND|MF_CHECKED);
	  CheckMenuItem(hMenu, IDM_POP6, MF_BYCOMMAND|MF_UNCHECKED);
	  CheckMenuItem(hMenu, IDM_POP5, MF_BYCOMMAND|MF_UNCHECKED);
      
	  zoomRate = 3;
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
       
	  break;

	case IDM_POP9:
	  hMenu = GetMenu(hWindow);
	  autoRefine = 0;
	  CheckMenuItem(hMenu, IDM_POP9, MF_BYCOMMAND|MF_CHECKED);
	  CheckMenuItem(hMenu, IDM_POP10, MF_BYCOMMAND|MF_UNCHECKED);
	  CheckMenuItem(hMenu, IDM_POP11, MF_BYCOMMAND|MF_UNCHECKED);
	  CheckMenuItem(hMenu, IDM_POP12, MF_BYCOMMAND|MF_UNCHECKED);
	  
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
	
	  break;
	
	case IDM_POP10:
	  hMenu = GetMenu(hWindow);
	  autoRefine ^= REFINE_INTERP;
	  if(autoRefine & REFINE_INTERP)
	    {
	      CheckMenuItem(hMenu, IDM_POP10, MF_BYCOMMAND|MF_CHECKED);
	      CheckMenuItem(hMenu, IDM_POP9, MF_BYCOMMAND|MF_UNCHECKED);
	      CheckMenuItem(hMenu, IDM_POP12, MF_BYCOMMAND|MF_UNCHECKED);
	      
	      if( autoRefine & REFINE_UV )
		{
		  autoRefine ^= REFINE_UV;
		}
	    }
	  else
	    {
	      CheckMenuItem(hMenu, IDM_POP10, MF_BYCOMMAND|MF_UNCHECKED);
	    }
	  if(autoRefine == 0)
	    CheckMenuItem(hMenu, IDM_POP9, MF_BYCOMMAND|MF_CHECKED);
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
	
	  break;
	  
	case IDM_POP11:
	  hMenu = GetMenu(hWindow);
	  autoRefine ^= REFINE_ZOOM;
	  if(autoRefine & REFINE_ZOOM)
	    {
	      CheckMenuItem(hMenu, IDM_POP11, MF_BYCOMMAND|MF_CHECKED);
	      CheckMenuItem(hMenu, IDM_POP9, MF_BYCOMMAND|MF_UNCHECKED);
	    }
	  else
	    {
	      CheckMenuItem(hMenu, IDM_POP11, MF_BYCOMMAND|MF_UNCHECKED);
	    }
	  if(autoRefine == 0)
	    CheckMenuItem(hMenu, IDM_POP9, MF_BYCOMMAND|MF_CHECKED);
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
       
	  break;

	case IDM_POP12:
	  hMenu = GetMenu(hWindow);
	  autoRefine ^= REFINE_UV;
	  if( autoRefine & REFINE_UV )
	    {
	      CheckMenuItem(hMenu, IDM_POP12, MF_BYCOMMAND|MF_CHECKED);
	      CheckMenuItem(hMenu, IDM_POP9, MF_BYCOMMAND|MF_UNCHECKED);
	      CheckMenuItem(hMenu, IDM_POP10, MF_BYCOMMAND|MF_UNCHECKED);
	      
	      if( autoRefine & REFINE_INTERP )
		{
		  autoRefine ^= REFINE_INTERP;
		}
	    }
	  else
	    {
	      CheckMenuItem(hMenu, IDM_POP12, MF_BYCOMMAND|MF_UNCHECKED);
	    }
	  if(autoRefine == 0)
	    CheckMenuItem(hMenu, IDM_POP9, MF_BYCOMMAND|MF_CHECKED);
	  drawPreviousFrame();
	  SendMessage(hWindow, wNewMessage, 0, 0);
     
	  break;

	case IDM_POP8:
	  PostQuitMessage(0);
	  break;
	}


       /* Set the draw_descr variable based on current state */
	  switch (interp) {
	  case 0:
	    draw_descr[1] = LF_NEAREST_UVST;
	    break;
	  case UVLERP:
	    draw_descr[1] = LF_BILINEAR_UV_NEAREST_ST;
	    break;
	  case STLERP:
	    draw_descr[1] = LF_NEAREST_UV_BILINEAR_ST;
	    break;
	  default:
	    draw_descr[1] = LF_QUADLINEAR_UVST;
	    break;
	  }

      break;
      
      /*need to repaint the window */
      
    case WM_PAINT:
   
      break;
    
    case WM_CREATE:
      globalWinHandle = hWindow;
      break;
      

      /*keyborad events */
    case WM_KEYUP:
      switch(wParam)
	{
	case VK_SHIFT:
	  SystemParametersInfo(SPI_SETKEYBOARDSPEED, keyboardDefaultSpeed, 0, NULL);
	  shift = 0;
	  /* redraw = 0; */
	  break;

	default:
	  SystemParametersInfo(SPI_SETKEYBOARDSPEED, keyboardDefaultSpeed, 0, NULL);
	  break;
	}
      break;

    case WM_KEYDOWN:
      switch(wParam)
	{
	case VK_SPACE:
	  SystemParametersInfo(SPI_SETKEYBOARDSPEED, 0, 0, NULL);
	  xobs=RESETXOBS;
	  yobs=RESETYOBS;
	  zobs=RESETZOBS;
	  xcenter = INITXCENTER;
	  ycenter = INITYCENTER;
	  zcenter = INITZCENTER;
	  break;

	case VK_UP:
	  if( fov*1.1 <= 180.0 )
	    {
	      fov *= 1.1;
	    }
	  break;

	case VK_DOWN:
	  fov /= 1.1;
	  break;
	  
	case VK_LEFT:
	  SystemParametersInfo(SPI_SETKEYBOARDSPEED, 0, 0, NULL);
	  single_frame = -1;
	  break;

	case VK_RIGHT:
	  SystemParametersInfo(SPI_SETKEYBOARDSPEED, 0, 0, NULL);
	  single_frame = 1;
	  break;
  
	case VK_SHIFT:
	  SystemParametersInfo(SPI_SETKEYBOARDSPEED, 0, 0, NULL);
	  shift = 1;
	  /* redraw = 0; */
	  break;

	case VK_ESCAPE:
	  done = 1;
	  PostQuitMessage(0);
	  break;

	case 'T':
	case 't':
	  SystemParametersInfo(SPI_SETKEYBOARDSPEED, 0, 0, NULL);
	  timeMode = 1-timeMode;
	  break;

	case 'f':
	case 'F':
	  SystemParametersInfo(SPI_SETKEYBOARDSPEED, 0, 0, NULL);
	  frameTimeMode = 1-frameTimeMode;
	  break;

	case 'm':
	case 'M':
	  SystemParametersInfo(SPI_SETKEYBOARDSPEED, 0, 0, NULL);
	  matrixTimeMode = 1-matrixTimeMode;
	  break;
	 
	default:
	  SystemParametersInfo(SPI_SETKEYBOARDSPEED, 0, 0, NULL);
	  break;
	}
      return;
      break;

      /*the mouse is moved */
    case WM_MOUSEMOVE:
      GetCursorPos(&pCursorNew);

      /*record old */
      pCursorOld = pCursorNew;
      if( (!leftdown)&&(!middledown)&&(!rightdown) )
	{
	  redraw = 0;
	  return;
	}
      break;
      
      /*the left mouse is pressed */
    case WM_LBUTTONDOWN:
      if( (!middledown) && (!rightdown) )
	{
	  SetCapture(hWindow);
	}
      leftdown= 1;
   
     
      break;
   
      /* the left mouse is released */
    case WM_LBUTTONUP:
      if( (!middledown) && (!rightdown) )
	{
	  ReleaseCapture();
	}
      leftdown = 0;
  
      break;
      
      /* the middle mouse is pressed */
    case WM_MBUTTONDOWN:
      if( (!leftdown) && (!rightdown) )
	{
	  SetCapture(hWindow);
	}
      middledown = 1;
      
      break;
   
      /* the middle mouse is released */
    case WM_MBUTTONUP:
      if( (!leftdown) && (!rightdown) )
	{
	  ReleaseCapture();
	}
      middledown = 0;
     
      break;
      
      /* the right mouse is pressed */
    case WM_RBUTTONDOWN:
      /* use right to simulate middle */
      if( (!leftdown) && (!middledown) )
	{
	  SetCapture(hWindow);
	}
      rightdown = 1;
      
      break;
   
      /* the right mouse is released */
    case WM_RBUTTONUP:
      /* use right to simulate middle */
      if( (!leftdown) && (!middledown) )
	{
	  ReleaseCapture();
	}
      rightdown = 0;
     
      break;

    default:
      times--;
      return DefWindowProc(hWindow, Message, wParam, lParam);
      
      break;
      }


  /* things to do after get the events */
  /*
  GetCursorPos(&pCursor);
  */
  dwMesPos = GetMessagePos();
  pCursor = MAKEPOINTS(dwMesPos);
  mx = ((float)pCursor.x-xorig)/xsize;
  my = ((float)pCursor.y-yorig)/ysize;
  dx = 2*(mx - lx);
  dy = -2*(my - ly);
  lx = mx;
  ly = my;


  if( Message != WM_MOUSEMOVE )
    {
      dx=0;
      dy=0;
    }

  switch(Message)
    {
    case WM_LBUTTONDOWN:
    case WM_MBUTTONDOWN:
    case WM_RBUTTONDOWN:
      return;
      break;
    default:
      break;
    }

 
  /* record the old */
  oldxobs = xobs;
  oldyobs = yobs;
  oldzobs = zobs;
  oldxcenter = xcenter;
  oldycenter = ycenter;
  oldzcenter = zcenter;

  /* Mouse interactions -- pasted from Marc's viewer */
  if (leftdown && (middledown || rightdown) ) {
    /* Mouse Y maps to zooming toward/from center
     * mouse X does nothing
     * Movement slows down as you approach the far plane
     * Don't let observer reach far plane unless multifield
     */
    

    if (dy < 0) {
      xobs = xobs+(xobs-xcenter)*dy;
      yobs = yobs+(yobs-ycenter)*dy;
      zobs = zobs+(zobs-zcenter)*dy;
    }
    else {
      xobs = xcenter+(xobs-xcenter)/(1-dy);
      yobs = ycenter+(yobs-ycenter)/(1-dy);
      zobs = zcenter+(zobs-zcenter)/(1-dy);
    }
    /* Restore dy to original sign */
    /* don't do this */
     dy = -dy; 
    
  }
  else if (leftdown) {
    /* Mouse XY maps to longitude and lattitude
     * as observer orbits center at fixed distance
     * If shifted, suppress lattitude changes,
     * easing flyarounds of full-surround fields
     */
    float aold,anew;
    rold = sqrt((xobs-xcenter)*(xobs-xcenter)+
		(zobs-zcenter)*(zobs-zcenter));
    aold = atan((xobs-xcenter)/(zobs-zcenter)) / PI*180.0;
    if (ABS(zobs-zcenter) < EPSILON) aold = 90.0 * 
				       SIGN(xobs-xcenter);
    else if (zobs<zcenter) aold = 180.0 + aold;
    anew = aold - dx * 90.0;
    
    xobs = xcenter + rold * sin(anew/180.0*PI);
    zobs = zcenter + rold * cos(anew/180.0*PI);
    
    rold = sqrt((xobs-xcenter)*(xobs-xcenter)+
		(yobs-ycenter)*(yobs-ycenter)+
		(zobs-zcenter)*(zobs-zcenter));
    newyobs = yobs-dy*rold;
    if (!shift) {
      yobs = newyobs;
    }
    rnew = sqrt((xobs-xcenter)*(xobs-xcenter)+
		(yobs-ycenter)*(yobs-ycenter)+
		(zobs-zcenter)*(zobs-zcenter));
    newxobs = (xobs-xcenter)*(rold/rnew)+xcenter;
    newyobs = (yobs-ycenter)*(rold/rnew)+ycenter;
    newzobs = (zobs-zcenter)*(rold/rnew)+zcenter;
    xobs = newxobs;
    if (!shift) {
      yobs = newyobs;
    }
    zobs = newzobs;
  }
  else if (middledown||rightdown) {
    /* Mouse XY maps to translation of viewed image      */
    /* achieved by matched moves of observer and center  */
    /* perpendicular to the ray connecting them          */
    /* Such translation shows off horizontal parallax   */
    if (ABS(zcenter-zobs) > ABS(xcenter-xobs)) {
      /* Front and back views */
      newxcenter = xcenter+(zcenter-zobs)*dx/2;
      newzcenter = zcenter-(xcenter-xobs)*dx/2;
      newycenter = ycenter-ABS(zcenter-zobs)*dy/2;
      newzcenter = newzcenter-(ycenter-yobs)*dy/2;
      newxobs = xobs+(zcenter-zobs)*dx/2;
      newzobs = zobs-(xcenter-xobs)*dx/2;
      newyobs = yobs-ABS(zcenter-zobs)*dy/2;
      newzobs = newzobs-(ycenter-yobs)*dy/2;
    }
    else {
      /* Side views */
      newzcenter = zcenter-(xcenter-xobs)*dx/2;
      newxcenter = xcenter+(zcenter-zobs)*dx/2;
      newycenter = ycenter-ABS(xcenter-xobs)*dy/2;
      newxcenter = newxcenter-(ycenter-yobs)*dy/2;
      newzobs = zobs-(xcenter-xobs)*dx/2;
      newxobs = xobs+(zcenter-zobs)*dx/2;
      newyobs = yobs-ABS(xcenter-xobs)*dy/2;
      newxobs = newxobs-(ycenter-yobs)*dy/2;
    }
    xcenter = newxcenter;
    ycenter = newycenter;
    zcenter = newzcenter;
    /* If shifted, movement of observer is suppressed, */
    /* causing mouse XY to map to viewing direction at */
    /* the (fixed) current position, which exhibits    */
    /* no horizontal parallax but is often useful      */
    if (!shift) {
      xobs = newxobs;
      yobs = newyobs;
      zobs = newzobs;
    } 
  }
  
  /* draw things here */
  if( redraw && (globalWinHandle != -1) ) 
    {
      lfBegin(LF_FIELD, NULL);
      if( (autoRefine & REFINE_INTERP) && !(middledown || leftdown || rightdown)
	  && (Message != WM_KEYDOWN) && !(updown || downdown) )
	{
	  oldlerp = draw_descr[1];
	  draw_descr[1] = LF_QUADLINEAR_UVST;
	  lfDraw(LF_COLOR, draw_descr, NULL);
	  draw_descr[1] = oldlerp;
	  whichCase = 1;
	}
      else if( (autoRefine & REFINE_UV) && !(middledown || leftdown || rightdown)
	  && (Message != WM_KEYDOWN) && !(updown || downdown) )
	{
	  oldlerp = draw_descr[1];
	  draw_descr[1] = LF_BILINEAR_UV_NEAREST_ST;
	  lfDraw(LF_COLOR, draw_descr, NULL);
	  draw_descr[1] = oldlerp;
	  whichCase = 2;
	}
      else
	{
	  lfDraw(LF_COLOR, draw_descr, NULL);
	  whichCase = 3;
	}
      lfEnd();
      lastDrawTime = GetCurrentTime();
      
      /* debug */
      /*
      sprintf(buffer, "old:%x, new:%x, (%f,%f,%f), (%f,%f,%f)", realOld, Message, xobs, yobs, zobs,
	      xcenter, ycenter, zcenter );
      pcdebug(0, 0, buffer);
      */
      /* end debug */
    }
  times--;
}



/* for debug */
int pcdebug(int x, int y, char * message)
{
  HDC hDC;
  
  hDC = GetDC(globalWinHandle);
  TextOut(hDC, x,y, message, strlen(message));
  ReleaseDC(globalWinHandle, hDC);
}
/* end debug */


/* stuff copied from Lucas's code */

void ParseArgs(int argc, char argv[][200])
{
  char *progName = argv[0];
  int i=0;

  argc--;
  argv++;

  while (argc > 0) {
    /* Verbose flag */
    if (!strcmp(argv[0], "-v")) {
      /*
      lfVerbose = 1;
      */
      argc--;
      argv++;
    } 
    /* size */
    else if (!strcmp(argv[i], "-size")) {
      if (argc < 3) PrintUsage(progName);
      xsize = atoi(argv[1]);
      ysize = atoi(argv[2]);
      /* error checking */
      if (xsize <2 || ysize < 2 || xsize > WEI_MAX_WINDOW_SIZE || 
	  ysize > WEI_MAX_WINDOW_SIZE) {
	lfError("Invalid window size... using defaults\n");
	xsize = WINDOW_WIDTH;
	ysize = WINDOW_HEIGHT;
      }
      argc -= 3;
      argv += 3;
    } 

    /* zoom */
    else if (!strcmp(argv[0], "-zoom")) {
      if (argc < 2) PrintUsage(progName);
      zoomRate = atoi(argv[1]);
      /* error checking */
      if (zoomRate < 1 || zoomRate > 100) {
	lfError("Invalid zoom scale... using default\n");
	zoomRate = 2.0;
      }
      argc -= 2;
      argv += 2;
    } 
    
    /* Interpolation mode */
    else if (!strcmp(argv[0], "-lerp")) {
      if (argc < 2) PrintUsage(progName);
      if (!strcmp(argv[1], "none")) {
	interp = 0;
      } else if (!strcmp(argv[1], "uv")) {
	interp = UVLERP;
      } else if (!strcmp(argv[1], "st")) {
	interp = STLERP;
      } else if (!strcmp(argv[1], "uvst")) {
	interp = STLERP + UVLERP;
      } else {
	lfError("Invalid lerp mode: %s\n", argv[1]);
	PrintUsage(progName);
      }
      argc -= 2;
      argv += 2;
    }

    /* Auto-Refine */
    else if (!strcmp(argv[0], "-refine")) {
      if (argc < 2) PrintUsage(progName);
      if (!strcmp(argv[1], "all")) {
	autoRefine = REFINE_INTERP + REFINE_ZOOM;
      } else if (!strcmp(argv[1], "uvst")) {
	autoRefine = REFINE_INTERP;
      } else if (!strcmp(argv[1], "zoom")) {
	autoRefine = REFINE_ZOOM;
      } else if (!strcmp(argv[1], "off")) {
	autoRefine = 0;
      } else {
	lfError("Invalid auto-refine mode: %s\n", argv[1]);
	PrintUsage(progName);
      }
      argc -= 2;
      argv += 2;
    }
    /* filename */
    else if (strncmp(argv[0], "-", 1)) {
      if (basename == NULL) {
	basename = argv[0];
      } else {
	lfError("Ignoring extra file %s\n", basename);
      } 
      argc--;
      argv++;
    } 
    
    /* default */
    else {
      PrintUsage(progName);
      exit(-1);
    }
  }

  /* Check to make sure basename has been set */
  if (basename == NULL) {
    PrintUsage(progName);
  }
}
  

void PrintUsage(char *programName)
{
  fprintf(stderr, 
	  "Usage: %s [-v] [-size <x> <y>] [-zoom <z>] [-lerp <mode>] \n"
	  "               [-refine on/off] <lightfield>\n"
	  "\n"
	  "     -v              Verbose on\n"
	  "\n"
	  "     -size <x> <y>   Size (in pixels) of display window\n"
	  "\n"
	  "     -zoom <z>       Zoom factor for display window\n"
	  "                     (The actual window size will be x*z, y*z)\n"
	  "\n"
	  "     -lerp uvst      Start with full uvst interpolation filter\n"
	  "     -lerp uv        Start with only uv interpolation\n"
	  "     -lerp st        Start with only st interpolation\n"
	  "     -lerp none      Start with point sampling\n"
	  "\n"
	  "     -refine on      Auto-refine image when viewer isn't moving\n"
	  "     -refine off     Never auto-refine image\n"
	  "\n"
	  "     <lightfield>    This can be the name of either a .lif\n"
	  "                     file or a .lid file.  If it is a .lid\n"
	  "                     file, it will tell lifview where to\n"
	  "                     find the raw .rgb images.\n"
	  "\n"
	  , programName);
  
  exit(-1);
}


void handleCompressedFile(char *fileName)
{
  static char buffer[200];
  
  system("del temp.gz");
  sprintf(buffer, "copy %s temp.gz", fileName);
  system(buffer);
  sprintf(buffer, "gzip -d -c temp.gz >temp.lif");
  system(buffer);
  system("del temp.gz");
}


LF_FileType DetermineFileType(char *fileName)
{
  LF_FileType type = LF_UNKNOWN;
  char buffer[20];
  FILE *inFile = NULL;
  char *suffixPtr, *tPtr;

  /* Open the input file */
#ifdef sgi
  inFile = fopen(fileName, "rb");
#else /* WIN32 */
  inFile = fopen(fileName, "r+b");
#endif 
  if (!inFile) {
    lfError("Cannot open file %s\n", fileName);
    exit(-1);
  }
  
  /* Get the first 8 characters of the file to determine its type */
  fgets(buffer, 9, inFile);
  
  switch(buffer[0]) {
  case 'L':  /* LIF1.0 */
    if (!strncmp(buffer, "LIF", 3)) {
      if (!strncmp(buffer, "LIF1.0", 6)) {
	type = LF_LIF;
      } else {
	lfError("Cannot handle LIF version number \"%s\".\n"
		"This is lifview 1.0.\n", &(buffer[3]));
	type = LF_LIFBADVER;
      }
      
    } else {
      type = LF_UNKNOWN;
    }
    break;
  case '#': /* # LIF1.0 */
    type = LF_LID;
    break;

  case 31:  /* gzip magic byte */
    type = LF_GZ;
    break;
  default:
    type = LF_UNKNOWN;
    break;
  }

  /* If the type is still unknown, look at the suffix of the
   * filename, and try to use it if possible, or at least print
   * a reasonable error message. 
   */
  if (type == LF_UNKNOWN) {
    /* Figure out the suffix */
    suffixPtr = fileName;
    /* Make suffixPtr point to character after the last . */
    for (tPtr = fileName; tPtr < (fileName + strlen(fileName)); tPtr++) {
      if (*tPtr == '.') suffixPtr = tPtr+1;
    }
    
    /* Now test it against known types */
    if (!strcmp(".lif", suffixPtr) || !strcmp(".LIF", suffixPtr)) {
      lfError("Bad .lif file, does not start with \"LIF1.0\".\n");
    }
    else if (!strcmp(".gz", suffixPtr) || !strcmp(".GZ", suffixPtr)) {
      lfError("Bad .gz file, does not start with magic characters.\n");
    }
    else if (!strcmp(".lid", suffixPtr) || !strcmp(".LID", suffixPtr)) {
      /* This is ok, I guess.  They might've stripped out the comments. */
      type = LF_LID;
    } else {
      lfError("Unrecognized input format for file \"%s\".\n", fileName);
    }
  }

  /* Close up file, return */
  fclose(inFile);
  return(type);
}
